import logging

log = logging.getLogger(__name__)


def valid_field_ids(field_datasets, fields=None):
    """
    Parameters
    ----------
    field_datasets : list of :class:`daskms.Dataset`

    fields : None or str
        if None, all fields are valid, otherwise
        the string will be split on commas and used to
        select fields. Both field ID's and names are supported
        in the string.

        .. code-block:: python

            valid_field_ids(field_ds, ["PKS-1934,1,DEEP2,3"])

    Returns
    -------
    fields_ids : list of ints
        List of valid field IDs
    """
    if fields is None:
        return list(range(len(field_datasets)))
    elif not isinstance(fields, str):
        raise TypeError("fields must be None or str")

    field_names = {}

    for i, ds in enumerate(field_datasets):
        names = ds.NAME.values
        assert len(names) == 1, "Should have a single name per row"
        field_names[str(names[0])] = i

    selected_fields = [f.strip() for f in fields.split(',')]

    # Add integral lookup values
    for field_name, field_id in list(field_names.items()):
        exists = field_names.setdefault(str(field_id), field_id)

        if not exists == field_id:
            raise ValueError("Existing field_id mismatch "
                             "for field %s. %d != %d"
                             % (field_name, exists, field_id))

    field_ids = []

    for f in selected_fields:
        try:
            field_id = field_names[f]
        except KeyError:
            log.warning("Field %s does not exist and ignored for selection", f)
        else:
            field_ids.append(field_id)

    return field_ids


def filter_datasets(datasets, field_ids):
    """
    Parameters
    ----------
    datasets : list of :class:`daskms.Dataset`
        List of datasets representing partitions of the Measurement Set.
        Should be partitioned by FIELD_ID and DATA_DESC_ID at least
    field_ids : list of ints
        List of valid fields

    Returns
    -------
    datasets : list of :class:`daskms.Dataset`
        Filtered list of datasets
    """
    field_ids = set(field_ids)
    return [ds for ds in datasets if ds.FIELD_ID in field_ids]



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
from contextlib import ExitStack

from dask.diagnostics import ProgressBar
import numpy as np
import sys


try:
    import dask
    import dask.array as da
    from daskms import xds_from_ms, xds_from_table, xds_to_table
except ImportError as e:
    opt_import_error = e
else:
    opt_import_error = None

from africanus.coordinates.dask import radec_to_lm
from africanus.rime.dask import phase_delay, predict_vis
from africanus.model.coherency.dask import convert
from africanus.model.shape.dask import gaussian
from africanus.util.requirements import requires_optional

from Crystalball.budget import get_budget
from Crystalball.filtering import valid_field_ids, filter_datasets
from Crystalball.ms import ms_preprocess
from Crystalball.region import load_regions
from Crystalball.wsclean import import_from_wsclean


def create_parser():
    p = argparse.ArgumentParser()
    p.add_argument("ms",
                   help="Input .MS file.")
    p.add_argument("-sm", "--sky-model", default="sky-model.txt",
                   help="Name of file containing the sky model. "
                        "Default is 'sky-model.txt'")
    p.add_argument("-o", "--output-column", default="MODEL_DATA",
                   help="Output visibility column. Default is '%(default)s'")
    p.add_argument("-f", "--fields", type=str,
                   help="Comma-separated list of Field names or ids "
                        "which should be predicted. "
                        "All fields are predicted by default.")
    p.add_argument("-rc", "--row-chunks", type=int, default=0,
                   help="Number of rows of input .MS that are processed in "
                        "a single chunk. If 0 it will be set automatically. "
                        "Default is 0.")
    p.add_argument("-mc", "--model-chunks", type=int, default=0,
                   help="Number of sky model components that are processed in "
                        "a single chunk. If 0 it wil be set automatically. "
                        "Default is 0.")
    p.add_argument("--exp-sign-convention", choices=['casa', 'thompson'],
                   default='casa',
                   help="Sign convention to use for the complex exponential. "
                        "'casa' specifies the e^(2.pi.I) convention while "
                        "'thompson' specifies the e^(-2.pi.I) convention in "
                        "the white book and Fourier analysis literature. "
                        "Defaults to '%(default)s'")
    p.add_argument("-sp", "--spectra", action="store_true",
                   help="Optional. Model sources as non-flat spectra. "
                        "The spectral coefficients and reference frequency "
                        "must be present in the sky model.")
    p.add_argument("-w", "--within", type=str,
                   help="Optional. Give JS9 region file. Only sources within "
                        "those regions will be included.")
    p.add_argument("-po", "--points-only", action="store_true",
                   help="Select only point-type sources.")
    p.add_argument("-ns", "--num-sources", type=int, default=0, metavar="N",
                   help="Select only N brightest sources.")
    p.add_argument("-j", "--num-workers", type=int, default=0, metavar="N",
                   help="Explicitly set the number of worker threads.")
    p.add_argument("-mf", "--memory-fraction", type=float, default=0.5,
                   help="Fraction of system RAM that can be used. "
                        "Used when setting automatically the "
                        "chunk size. Default in 0.5.")

    return p


def support_tables(args, tables):
    """
    Parameters
    ----------
    args : object
        Script argument objects
    tables : list of str
        List of support tables to open

    Returns
    -------
    table_map : dict of :class:`xarray.Dataset`
        {name: dataset}
    """
    return {t: [ds.compute() for ds in
                xds_from_table("::".join((args.ms, t)),
                               group_cols="__row__")]
            for t in tables}


def corr_schema(pol):
    """
    Parameters
    ----------
    pol : :class:`xarray.Dataset`

    Returns
    -------
    corr_schema : list of list
        correlation schema from the POLARIZATION table,
        `[[9, 10], [11, 12]]` for example
    """

    corrs = pol.NUM_CORR.data[0]
    corr_types = pol.CORR_TYPE.data[0]

    if corrs == 4:
        return [[corr_types[0], corr_types[1]],
                [corr_types[2], corr_types[3]]]  # (2, 2) shape
    elif corrs == 2:
        return [corr_types[0], corr_types[1]]    # (2, ) shape
    elif corrs == 1:
        return [corr_types[0]]                   # (1, ) shape
    else:
        raise ValueError("corrs %d not in (1, 2, 4)" % corrs)


def einsum_schema(pol, dospec):
    """
    Returns an einsum schema suitable for multiplying per-baseline
    phase and brightness terms.

    Parameters
    ----------
    pol : :class:`xarray.Dataset`

    Returns
    -------
    einsum_schema : str
    """
    corrs = pol.NUM_CORR.values

    if corrs == 4:
        if dospec:
            return "srf, sfij -> srfij"
        else:
            return "srf, sij -> srfij"
    elif corrs in (2, 1):
        if dospec:
            return "srf, sfi -> srfi"
        else:
            return "srf, si -> srfi"
    else:
        raise ValueError("corrs %d not in (1, 2, 4)" % corrs)


def predict():
    # Parse application args
    args = create_parser().parse_args([a for a in sys.argv[1:]])

    with ExitStack() as stack:
        # Set up dask ThreadPool prior to any application dask calls
        if args.num_workers:
            stack.enter_context(dask.config.set(num_workers=args.num_workers))

        # Run application script
        return _predict(args)


@requires_optional("dask.array", "daskms", opt_import_error)
def _predict(args):
    # get inclusion regions
    include_regions = load_regions(args.within) if args.within else []

    # Import source data from WSClean component list
    # See https://sourceforge.net/p/wsclean/wiki/ComponentList
    (comp_type, radec, stokes,
     spec_coeff, ref_freq, log_spec_ind,
     gaussian_shape) = import_from_wsclean(args.sky_model,
                                           include_regions=include_regions,
                                           point_only=args.points_only,
                                           num=args.num_sources or None)

    # Add output column if it isn't present
    ms_rows, ms_datatype = ms_preprocess(args)

    # Get the support tables
    tables = support_tables(args, ["FIELD", "DATA_DESCRIPTION",
                                   "SPECTRAL_WINDOW", "POLARIZATION"])

    field_ds = tables["FIELD"]
    ddid_ds = tables["DATA_DESCRIPTION"]
    spw_ds = tables["SPECTRAL_WINDOW"]
    pol_ds = tables["POLARIZATION"]

    max_num_chan = max([ss.NUM_CHAN.data[0] for ss in spw_ds])
    max_num_corr = max([ss.NUM_CORR.data[0] for ss in pol_ds])

    # Perform resource budgeting
    args.row_chunks, args.model_chunks = get_budget(comp_type.shape[0],
                                                    ms_rows,
                                                    max_num_chan,
                                                    max_num_corr,
                                                    ms_datatype, args)

    radec = da.from_array(radec, chunks=(args.model_chunks, 2))
    stokes = da.from_array(stokes, chunks=(args.model_chunks, 4))

    if np.count_nonzero(comp_type == 'GAUSSIAN') > 0:
        gaussian_components = True
        gshape_chunks = (args.model_chunks, 3)
        gaussian_shape = da.from_array(gaussian_shape, chunks=gshape_chunks)
    else:
        gaussian_components = False

    if args.spectra:
        spec_chunks = (args.model_chunks, spec_coeff.shape[1])
        spec_coeff = da.from_array(spec_coeff, chunks=spec_chunks)
        ref_freq = da.from_array(ref_freq, chunks=(args.model_chunks,))

    # List of write operations
    writes = []

    # Construct a graph for each FIELD and DATA DESCRIPTOR
    datasets = xds_from_ms(args.ms,
                           columns=["UVW", "ANTENNA1", "ANTENNA2", "TIME"],
                           group_cols=["FIELD_ID", "DATA_DESC_ID"],
                           chunks={"row": args.row_chunks})

    select_fields = valid_field_ids(field_ds, args.fields)

    for xds in filter_datasets(datasets, select_fields):
        # Extract frequencies from the spectral window associated
        # with this data descriptor id
        field = field_ds[xds.attrs['FIELD_ID']]
        ddid = ddid_ds[xds.attrs['DATA_DESC_ID']]
        spw = spw_ds[ddid.SPECTRAL_WINDOW_ID.data[0]]
        pol = pol_ds[ddid.POLARIZATION_ID.data[0]]
        frequency = spw.CHAN_FREQ.data[0]

        corrs = pol.NUM_CORR.values

        lm = radec_to_lm(radec, field.PHASE_DIR.data[0][0])

        if args.exp_sign_convention == 'casa':
            uvw = -xds.UVW.data
        elif args.exp_sign_convention == 'thompson':
            uvw = xds.UVW.data
        else:
            raise ValueError("Invalid sign convention '%s'" % args.sign)

        if args.spectra:
            # flux density at reference frequency ...
            # ... for logarithmic polynomial functions
            if log_spec_ind:
                Is = da.log(stokes[:, 0, None])*frequency[None, :]**0
            # ... or for ordinary polynomial functions
            else:
                Is = stokes[:, 0, None]*frequency[None, :]**0
            # additional terms of SED ...
            for jj in range(spec_coeff.shape[1]):
                # ... for logarithmic polynomial functions
                if log_spec_ind:
                    Is += spec_coeff[:, jj, None] * \
                        da.log((frequency[None, :]/ref_freq[:, None])**(jj+1))
                # ... or for ordinary polynomial functions
                else:
                    Is += spec_coeff[:, jj, None] * \
                        (frequency[None, :]/ref_freq[:, None]-1)**(jj+1)
            if log_spec_ind:
                Is = da.exp(Is)
            Qs = da.zeros_like(Is)
            Us = da.zeros_like(Is)
            Vs = da.zeros_like(Is)
            # stack along new axis and make it the last axis of the new array
            spectrum = da.stack([Is, Qs, Us, Vs], axis=-1)
            spectrum = spectrum.rechunk(
                spectrum.chunks[:2] + (spectrum.shape[2],))

        print('-------------------------------------------')
        print('Nr sources        = {0:d}'.format(stokes.shape[0]))
        print('-------------------------------------------')
        print('stokes.shape      = {0:}'.format(stokes.shape))
        print('frequency.shape   = {0:}'.format(frequency.shape))
        if args.spectra:
            print('Is.shape          = {0:}'.format(Is.shape))
        if args.spectra:
            print('spectrum.shape    = {0:}'.format(spectrum.shape))

        # (source, row, frequency)
        phase = phase_delay(lm, uvw, frequency)
        # If at least one Gaussian component is present in the component
        # list then all sources are modelled as Gaussian components
        # (Delta components have zero width)
        if gaussian_components:
            phase *= gaussian(uvw, frequency, gaussian_shape)
        # (source, frequency, corr_products)
        brightness = convert(spectrum if args.spectra else stokes,
                             ["I", "Q", "U", "V"],
                             corr_schema(pol))

        print('brightness.shape  = {0:}'.format(brightness.shape))
        print('phase.shape       = {0:}'.format(phase.shape))
        print('-------------------------------------------')
        print('Attempting phase-brightness einsum with "{0:s}"'
              .format(einsum_schema(pol, args.spectra)))

        # (source, row, frequency, corr_products)
        jones = da.einsum(einsum_schema(pol, args.spectra), phase, brightness)
        print('jones.shape       = {0:}'.format(jones.shape))
        print('-------------------------------------------')
        if gaussian_components:
            print('Some Gaussian sources found')
        else:
            print('All sources are Delta functions')
        print('-------------------------------------------')

        # Identify time indices
        _, time_index = da.unique(xds.TIME.data, return_inverse=True)

        # Predict visibilities
        vis = predict_vis(time_index, xds.ANTENNA1.data, xds.ANTENNA2.data,
                          None, jones, None, None, None, None)

        # Reshape (2, 2) correlation to shape (4,)
        if corrs == 4:
            vis = vis.reshape(vis.shape[:2] + (4,))

        # Assign visibilities to MODEL_DATA array on the dataset
        xds = xds.assign(
            **{args.output_column: (("row", "chan", "corr"), vis)})
        # Create a write to the table
        write = xds_to_table(xds, args.ms, [args.output_column])
        # Add to the list of writes
        writes.append(write)

    with ExitStack() as stack:
        if sys.stdout.isatty():
            # Default progress bar in user terminal
            stack.enter_context(ProgressBar())
        else:
            # Log progress every 5 minutes
            stack.enter_context(ProgressBar(minimum=2*60, dt=5))

        # Submit all graph computations in parallel
        dask.compute(writes)



# -*- coding: utf-8 -*-

import logging

import pyrap.tables as pt

log = logging.getLogger(__name__)


def ms_preprocess(args):
    """
    Adds output column if missing.

    Returns
    -------
    ms_rows : int
        number of Measurement Set Rows
    """

    # check output column
    with pt.table(args.ms, readonly=False) as ms:
        # Return if nothing todo
        if args.output_column in ms.colnames():
            return ms.nrows(), ms.coldatatype('DATA')

        log.info('inserting new column %s', args.output_column)
        desc = ms.getcoldesc("DATA")
        desc['name'] = args.output_column
        # python version hates spaces, who knows why
        desc['comment'] = desc['comment'].replace(" ", "_")
        dminfo = ms.getdminfo("DATA")
        dminfo["NAME"] = "%s-%s" % (dminfo["NAME"], args.output_column)
        ms.addcols(desc, dminfo)

        return ms.nrows(), ms.coldatatype('DATA')



import psutil
import numpy as np


def get_budget(nr_sources, nr_rows, nr_chans, nr_corrs, data_type, cb_args,
               fudge_factor=2., row2source_ratio=100):
    systmem = np.float(psutil.virtual_memory()[0])
    if not cb_args.num_workers:
        nrthreads = psutil.cpu_count()
    else:
        nrthreads = cb_args.num_workers

    print('-------------------------------------------')
    print('system RAM = {0:.2f} GB'.format(systmem/1024**3))
    print('nr of logical CPUs = {0:d}'.format(nrthreads))
    print('nr sources = {0:d}'.format(nr_sources))
    print('nr rows    = {0:d}'.format(nr_rows))
    print('nr chans   = {0:d}'.format(nr_chans))
    print('nr corrs   = {0:d}'.format(nr_corrs))

    data_type = {'complex': 'complex64', 'dcomplex': 'complex128'}[data_type]
    data_bytes = np.dtype(data_type).itemsize
    bytes_per_row_source = nr_chans*nr_corrs*data_bytes
    memory_per_row_source = bytes_per_row_source * fudge_factor

    if cb_args.model_chunks and cb_args.row_chunks:
        rows_per_chunk = cb_args.row_chunks
        sources_per_chunk = cb_args.model_chunks
        print('sources per chunk = {0:.0f} (user setting)'.format(
            sources_per_chunk))
        print('rows per chunk    = {0:.0f} (user setting)'.format(
            rows_per_chunk))
        memory_usage = (rows_per_chunk * sources_per_chunk *
                        memory_per_row_source * nrthreads)
        print('expected memory usage = {0:.2f} GB'.format(
            memory_usage/1024**3))
    elif not cb_args.model_chunks and not cb_args.row_chunks:
        allowed_rowXsource_per_thread = systmem * \
            cb_args.memory_fraction / memory_per_row_source / nrthreads
        rows_per_chunk = np.int(np.minimum(nr_rows, np.sqrt(
            allowed_rowXsource_per_thread * row2source_ratio)))
        sources_per_chunk = np.int(np.minimum(
            nr_sources, rows_per_chunk / row2source_ratio))
        print('sources per chunk = {0:.0f} (auto setting)'.format(
            sources_per_chunk))
        print('rows per chunk    = {0:.0f} (auto setting)'.format(
            rows_per_chunk))
        memory_usage = (rows_per_chunk * sources_per_chunk *
                        memory_per_row_source*nrthreads)
        print('expected memory usage = {0:.2f} GB'.format(
            memory_usage/1024**3))
    else:
        raise ValueError('For now you must set both row and source chunk, or '
                         'leave both unset (=0); '
                         'you cannot set only one of them.')

    return rows_per_chunk, sources_per_chunk



#!/usr/bin/env python

from distutils.core import setup

requirements = [
    "codex-africanus[dask] >= 0.2.0",
    "dask-ms >= 0.2.0",
    "regions",
    "psutil"
]

extras_require = {
  'testing': ['pytest', 'pytest-flake8']
}

PACKAGE_NAME = 'crystalball'
__version__ = '0.2.2'

setup(name=PACKAGE_NAME,
      version=__version__,
      description="Predicts visibilities from a parameterised sky model",
      author="Paolo Serra",
      author_email="paolo80serra@gmail.com",
      entry_points={
        'console_scripts': ['crystalball=Crystalball.crystalball:predict'],
      },
      extras_require=extras_require,
      url="https://github.com/paoloserra/crystalball",
      packages=["Crystalball"],
      install_requires=requirements,
      include_package_data=True,
      python_requires=">=3.5",
      license=["GNU GPL v2"],
      classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Astronomy"
      ])
